@@ -63,15 +63,13 @@ class BokehForLabeledText(ABC):
         self._setup_dfs(df_dict)
         self._setup_sources()
         self._activate_search_builtin()
+        self.figure = figure(**self.figure_kwargs)
         self.reset_figure()
 
     def reset_figure(self):
         """Start over on the figure."""
-        logger.info("Creating/resetting Figure")
-        if hasattr(self, "figure"):
-            self.figure.renderers.clear()
-        else:
-            self.figure = figure(**self.figure_kwargs)
+        logger.info("Initializing/resetting Figure")
+        self.figure.renderers.clear()
 
     def _setup_widgets(self):
         """
@@ -91,17 +89,17 @@ class BokehForLabeledText(ABC):
             title="Text does not contain:", width_policy="fit", height_policy="fit"
         )
 
-    def layout_widgets(self):
+    def _layout_widgets(self):
         """Define the layout of widgets."""
         return column(self.search_pos, self.search_neg)
 
     def view(self):
         """Define the layout of the whole explorer."""
-        return column(self.layout_widgets(), self.figure)
+        return column(self._layout_widgets(), self.figure)
 
-    def _setup_dfs(self, df_dict):
+    def _setup_dfs(self, df_dict, copy=False):
         """
-        Check and store DataFrames.
+        Check and store DataFrames BY REFERENCE BY DEFAULT.
 
         Intended to be extended in child classes for pre/post processing.
         """
@@ -111,7 +109,9 @@ class BokehForLabeledText(ABC):
             set(df_dict.keys()) == expected_keys
         ), f"Expected the keys of df_dict to be exactly {expected_keys}"
 
-        self.dfs = {_key: _df.copy() for _key, _df in df_dict.items()}
+        self.dfs = {
+            _key: (_df.copy() if copy else _df) for _key, _df in df_dict.items()
+        }
 
     def _setup_sources(self):
         """
@@ -122,6 +122,17 @@ class BokehForLabeledText(ABC):
         logger.info("Setting up sources")
         self.sources = {_key: ColumnDataSource(_df) for _key, _df in self.dfs.items()}
 
+    def _update_sources(self):
+        """
+        Update the sources with the corresponding dfs.
+
+        Note that it seems mandatory to re-activate the search widgets.
+        This is because the source loses plotting kwargs.
+        """
+        for _key in self.__class__.DATA_KEY_TO_KWARGS.keys():
+            self.sources[_key].data = self.dfs[_key]
+        self._activate_search_builtin()
+
     def _activate_search_builtin(self):
         """
         Typically called once during initialization.
@@ -223,6 +234,35 @@ class BokehForLabeledText(ABC):
         self.search_neg.js_on_change("value", search_callback)
         return updated_kwargs
 
+    def _prelink_check(self, other):
+        """
+        Sanity check before linking two explorers.
+        """
+        assert other is not self, "Self-loops are fordidden"
+        assert isinstance(other, BokehForLabeledText), "Must link to BokehForLabelText"
+
+    def link_selection(self, key, other, other_key):
+        """
+        Sync the selected indices between specified sources.
+        """
+        self._prelink_check(other)
+        # link selection in a bidirectional manner
+        sl, sr = self.sources[key], other.sources[other_key]
+        sl.selected.js_link("indices", sl.selected, "indices")
+        sr.selected.js_link("indices", sr.selected, "indices")
+
+    def link_xy_range(self, other):
+        """
+        Sync plotting ranges on the xy-plane.
+        """
+        self._prelink_check(other)
+        # link coordinate ranges in a bidirectional manner
+        for _attr in ["start", "end"]:
+            self.figure.x_range.js_link(_attr, other.figure.x_range, _attr)
+            self.figure.y_range.js_link(_attr, other.figure.y_range, _attr)
+            other.figure.x_range.js_link(_attr, self.figure.x_range, _attr)
+            other.figure.y_range.js_link(_attr, self.figure.y_range, _attr)
+
     @abstractmethod
     def plot(self, *args, **kwargs):
         """
@@ -256,13 +296,13 @@ class BokehCorpusExplorer(BokehForLabeledText):
         """
         super().__init__(df_dict, **kwargs)
 
-    def _setup_dfs(self, df_dict):
+    def _setup_dfs(self, df_dict, **kwargs):
         """Extending from the parent method."""
         for _key in self.__class__.DATA_KEY_TO_KWARGS.keys():
             for _col in ["text", "x", "y"]:
                 assert _col in df_dict[_key].columns
 
-        super()._setup_dfs(df_dict)
+        super()._setup_dfs(df_dict, **kwargs)
 
     def plot(self, *args, **kwargs):
         """
@@ -294,23 +334,18 @@ class BokehCorpusAnnotator(BokehCorpusExplorer):
         """Conceptually the same as the parent method."""
         super().__init__(df_dict, **kwargs)
 
-    def _setup_dfs(self, df_dict):
+    def _setup_dfs(self, df_dict, **kwargs):
         """
         Extending from the parent method.
 
         Add a "label" column if it is not present.
         """
-        super()._setup_dfs(df_dict)
+        super()._setup_dfs(df_dict, **kwargs)
 
         if not "label" in self.dfs["raw"].columns:
             self.dfs["raw"]["label"] = module_config.ABSTAIN_DECODED
 
-    def update_source(self):
-        """Note that it seems required to re-activate the search widgets."""
-        self.sources["raw"].data = self.dfs["raw"]
-        self._activate_search_builtin()
-
-    def layout_widgets(self):
+    def _layout_widgets(self):
         """Define the layout of widgets."""
         first_row = row(self.search_pos, self.search_neg)
         second_row = row(
@@ -323,7 +358,6 @@ class BokehCorpusAnnotator(BokehCorpusExplorer):
         Create annotator widgets and assign Python callbacks.
         """
         from bokeh.models import TextInput, Button
-        from bokeh.events import ButtonClick
 
         super()._setup_widgets()
 
@@ -341,7 +375,7 @@ class BokehCorpusAnnotator(BokehCorpusExplorer):
             width_policy="min",
         )
 
-        def apply():
+        def callback_apply():
             """
             A callback on clicking the 'self.annotator_apply' button.
 
@@ -357,27 +391,36 @@ class BokehCorpusAnnotator(BokehCorpusExplorer):
             example_new = self.dfs["raw"].at[selected_idx[0], "label"]
             logger.good(f"Updated DataFrame, e.g. {example_old} -> {example_new}")
 
-            self.update_source()
+            self._update_sources()
             self.plot()
             logger.good(f"Updated annotator plot at {current_time()}")
 
-        def export():
+        def callback_export(path=None):
             """
             A callback on clicking the 'self.annotator_export' button.
 
             Saves the dataframe to a pickle.
             """
             from dill import dump
-            from datetime import datetime
 
-            timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
-            filename = f"bokeh-annotated-df-{timestamp}.pkl"
-            with open(filename, "wb") as f:
+            # auto-determine the export path
+            if path is None:
+                timestamp = current_time("%Y%m%d%H%M%S")
+                path = f".bokeh-annotated-df-{timestamp}.pkl"
+
+            # save a pickle, then send a message
+            # note that excel/csv can be problematic with certain kinds of data
+            with open(path, "wb") as f:
                 dump(self.dfs["raw"], f)
-            logger.good(f"Saved DataFrame to {filename}")
+            logger.good(f"Saved DataFrame to {path}")
 
-        self.annotator_apply.on_event(ButtonClick, apply)
-        self.annotator_export.on_event(ButtonClick, export)
+        # keep the references to the callbacks
+        self._callback_apply = callback_apply
+        self._callback_export = callback_export
+
+        # assign callbacks
+        self.annotator_apply.on_click(self._callback_apply)
+        self.annotator_export.on_click(self._callback_export)
 
     def plot(self):
         """
@@ -427,12 +470,12 @@ class BokehMarginExplorer(BokehCorpusExplorer):
         self.label_col_b = label_col_b
         super().__init__(df_dict, **kwargs)
 
-    def _setup_dfs(self, df_dict):
+    def _setup_dfs(self, df_dict, **kwargs):
         """Extending from the parent method."""
         for _key in [self.label_col_a, self.label_col_b]:
             assert _key in df_dict["raw"].columns
 
-        super()._setup_dfs(df_dict)
+        super()._setup_dfs(df_dict, **kwargs)
 
     def plot(self, label, **kwargs):
         """
@@ -508,9 +551,9 @@ class BokehSnorkelExplorer(BokehCorpusExplorer):
         self.lfs = []
         self.palette = Category20[20]
 
-    def _setup_dfs(self, df_dict):
+    def _setup_dfs(self, df_dict, **kwargs):
         """Extending from the parent method."""
-        super()._setup_dfs(df_dict)
+        super()._setup_dfs(df_dict, **kwargs)
 
         assert "label" in self.dfs["labeled"].columns
         if not "label" in self.dfs["raw"].columns:
@@ -522,10 +565,10 @@ class BokehSnorkelExplorer(BokehCorpusExplorer):
         """
         Plot a single labeling function.
 
-        :param lf: labeling function decorated by @hover.utils.snorkel_helper.labeling_function()
-        :param L_raw: labeling function predictions, in decoded labels, on the raw df.
-        :param L_labeled: labeling function predictions, in decoded labels, on the labeled df.
-        :param include: subsets to show, which can be correct(C)/incorrect(I)/missed(M)/hit(H).
+        - param lf: labeling function decorated by @hover.utils.snorkel_helper.labeling_function()
+        - param L_raw: labeling function predictions, in decoded labels, on the raw df.
+        - param L_labeled: labeling function predictions, in decoded labels, on the labeled df.
+        - param include: subsets to show, which can be correct(C)/incorrect(I)/missed(M)/hit(H).
         """
         # keep track of added LF
         self.lfs.append(lf)
